import unittest
from unittest import mock


@mock.patch("deltacat.io.memcached_object_store.cloudpickle")
@mock.patch("deltacat.io.memcached_object_store.socket")
class TestMemcachedObjectStore(unittest.TestCase):

    TEST_VALUE = "test-value"

    def setUp(self):
        from deltacat.io.memcached_object_store import MemcachedObjectStore

        self.object_store = MemcachedObjectStore(
            storage_node_ips=["172.1.1.1", "172.2.2.2", "172.3.3.3"]
        )

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_put_many_sanity(
        self,
        mock_retrying_client,
        mock_client,
        mock_socket,
        mock_cloudpickle,
    ):
        mock_cloudpickle.dumps.return_value = self.TEST_VALUE
        mock_cloudpickle.loads.return_value = self.TEST_VALUE
        mock_socket.gethostbyname.return_value = "0.0.0.0"
        mock_socket.gethostname.return_value = "test-host"
        mock_retrying_client.return_value = mock_client.return_value
        mock_client.return_value.set_many.return_value = []

        result = self.object_store.put_many(["a", "b", "c"])

        self.assertEqual(3, len(result))
        self.assertRegex(result[0], ".*_.*")

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_put_many_when_cache_fails(
        self, mock_retrying_client, mock_client, mock_socket, mock_cloudpickle
    ):
        mock_cloudpickle.dumps.return_value = self.TEST_VALUE
        mock_cloudpickle.loads.return_value = self.TEST_VALUE
        mock_socket.gethostbyname.return_value = "0.0.0.0"
        mock_socket.gethostname.return_value = "test-host"
        mock_retrying_client.return_value = mock_client.return_value
        mock_client.return_value.set_many.return_value = ["abcd"]

        with self.assertRaises(RuntimeError):
            self.object_store.put_many(["a", "b"])

        self.assertEqual(1, mock_client.return_value.set_many.call_count)

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_get_many_sanity(
        self, mock_retrying_client, mock_client, mock_socket, mock_cloudpickle
    ):
        mock_cloudpickle.dumps.return_value = self.TEST_VALUE
        mock_cloudpickle.loads.return_value = self.TEST_VALUE
        mock_socket.gethostbyname.return_value = "0.0.0.0"
        mock_socket.gethostname.return_value = "test-host"
        mock_client.return_value.get_many.return_value = {
            "key1": "value1",
            "key2": "value2",
        }
        mock_retrying_client.return_value = mock_client.return_value

        result = self.object_store.get_many(["test_ip", "test_ip"])

        self.assertEqual(2, len(result))
        self.assertEqual(1, mock_client.return_value.get_many.call_count)

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_get_many_when_cache_expired(
        self, mock_retrying_client, mock_client, mock_socket, mock_cloudpickle
    ):
        mock_cloudpickle.dumps.return_value = self.TEST_VALUE
        mock_cloudpickle.loads.return_value = self.TEST_VALUE
        mock_socket.gethostbyname.return_value = "0.0.0.0"
        mock_socket.gethostname.return_value = "test-host"
        mock_client.return_value.get_many.return_value = {"key1": "value1"}
        mock_retrying_client.return_value = mock_client.return_value

        with self.assertRaises(AssertionError):
            self.object_store.get_many(["test_ip", "test_ip"])

        self.assertEqual(1, mock_client.return_value.get_many.call_count)

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_get_sanity(
        self, mock_retrying_client, mock_client, mock_socket, mock_cloudpickle
    ):
        mock_cloudpickle.dumps.return_value = self.TEST_VALUE
        mock_cloudpickle.loads.return_value = self.TEST_VALUE
        mock_socket.gethostbyname.return_value = "0.0.0.0"
        mock_socket.gethostname.return_value = "test-host"
        mock_client.return_value.get.return_value = self.TEST_VALUE
        mock_retrying_client.return_value = mock_client.return_value

        result = self.object_store.get("test_ip")

        self.assertEqual(self.TEST_VALUE, result)
        self.assertEqual(1, mock_client.return_value.get.call_count)

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_get_when_cache_fails(
        self, mock_retrying_client, mock_client, mock_socket, mock_cloudpickle
    ):
        mock_cloudpickle.dumps.return_value = self.TEST_VALUE
        mock_cloudpickle.loads.return_value = self.TEST_VALUE
        mock_socket.gethostbyname.return_value = "0.0.0.0"
        mock_socket.gethostname.return_value = "test-host"
        mock_client.return_value.get.side_effect = RuntimeError()
        mock_retrying_client.return_value = mock_client.return_value

        with self.assertRaises(RuntimeError):
            self.object_store.get("test_ip")

        self.assertEqual(1, mock_client.return_value.get.call_count)

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_put_sanity(
        self, mock_retrying_client, mock_client, mock_socket, mock_cloudpickle
    ):
        mock_cloudpickle.dumps.return_value = self.TEST_VALUE
        mock_cloudpickle.loads.return_value = self.TEST_VALUE
        mock_socket.gethostbyname.return_value = "0.0.0.0"
        mock_socket.gethostname.return_value = "test-host"
        mock_retrying_client.return_value = mock_client.return_value
        mock_client.return_value.set.return_value = True

        result = self.object_store.put("test")

        self.assertIsNotNone(result)
        self.assertEqual(1, mock_client.return_value.set.call_count)

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_put_when_cache_fails(
        self, mock_retrying_client, mock_client, mock_socket, mock_cloudpickle
    ):
        mock_cloudpickle.dumps.return_value = self.TEST_VALUE
        mock_cloudpickle.loads.return_value = self.TEST_VALUE
        mock_socket.gethostbyname.return_value = "0.0.0.0"
        mock_socket.gethostname.return_value = "test-host"
        mock_retrying_client.return_value = mock_client.return_value
        mock_client.return_value.set.return_value = False

        with self.assertRaises(RuntimeError):
            self.object_store.put("test_ip")

        self.assertEqual(1, mock_client.return_value.set.call_count)



from dataclasses import dataclass

import ray
import logging

from deltacat import logs
from enum import Enum
from typing import Dict, Any, List, Callable
from deltacat.aws.clients import resource_cache
from datetime import datetime

from ray._private.services import get_node_ip_address

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

DEFAULT_DELTACAT_METRICS_NAMESPACE = "ray-deltacat-metrics"


class MetricsTarget(str, Enum):
    CLOUDWATCH = "cloudwatch"


@dataclass
class MetricsConfig:
    def __init__(self, region: str, job_run_id: str, metrics_target: MetricsTarget):
        self.region = region
        self.job_run_id = job_run_id
        self.metrics_target = metrics_target


class MetricsType(str, Enum):
    TIMER = "timer"


class MetricsDimensionType(str, Enum):
    NODE_IP = "node_ip"
    RAY_TASK_ID = "task_id"
    RAY_WORKER_ID = "worker_id"


def _build_metrics_name(metrics_type: Enum, metrics_name: str) -> str:
    metrics_name_with_type = f"{metrics_name}_{metrics_type}"
    return metrics_name_with_type


def _get_current_node_ip() -> str:
    current_node_ip = get_node_ip_address()
    return {"Name": f"node_ip", "Value": f"{current_node_ip}"}


def _get_current_task_id() -> Dict[str, Any]:
    return {
        "Name": f"ray_task_id",
        "Value": f"{ray.get_runtime_context().get_task_id()}",
    }


def _get_current_worker_id() -> Dict[str, Any]:
    return {
        "Name": f"ray_worker_id",
        "Value": f"{ray.get_runtime_context().worker.core_worker.get_worker_id()}",
    }


METRICS_DIMENSION_TYPE_TO_VALUE_DICT: Dict[str, Callable] = {
    MetricsDimensionType.NODE_IP.value: _get_current_node_ip,
    MetricsDimensionType.RAY_TASK_ID.value: _get_current_task_id,
    MetricsDimensionType.RAY_WORKER_ID.value: _get_current_worker_id,
}


def _build_cloudwatch_metrics(
    metrics_name: str,
    metrics_type: Enum,
    value: str,
    dimension_types: List[Enum],
    timestamp: datetime,
    **kwargs,
) -> Dict[str, Any]:
    metrics_name_with_type = _build_metrics_name(metrics_type, metrics_name)
    dimensions = []
    for dimension_type in dimension_types:
        dimensions.append(
            METRICS_DIMENSION_TYPE_TO_VALUE_DICT.get(dimension_type.value)()
        )
    return [
        {
            "MetricName": f"{metrics_name_with_type}",
            "Dimensions": dimensions,
            "Timestamp": timestamp,
            "Value": value,
            **kwargs,
        }
    ]


def _emit_metrics(
    metrics_name: str,
    metrics_type: Enum,
    metrics_config: MetricsConfig,
    value: str,
    dimension_types: List[Enum],
    **kwargs,
) -> None:
    metrics_target = metrics_config.metrics_target
    assert isinstance(
        metrics_target, MetricsTarget
    ), f"{metrics_target} is not a valid supported metrics target type! "
    if metrics_target == MetricsTarget.CLOUDWATCH:
        _emit_cloudwatch_metrics(
            metrics_name=metrics_name,
            metrics_type=metrics_type,
            metrics_config=metrics_config,
            value=value,
            dimension_types=dimension_types,
            **kwargs,
        )
    else:
        logger.warning(f"{metrics_target} is not a supported metrics target type.")


def _emit_cloudwatch_metrics(
    metrics_name: str,
    metrics_type: Enum,
    metrics_config: MetricsConfig,
    value: str,
    dimension_types: List[Enum],
    **kwargs,
) -> None:
    ct = datetime.now()
    current_instance_region = metrics_config.region
    cloudwatch_resource = resource_cache("cloudwatch", current_instance_region)
    cloudwatch_client = cloudwatch_resource.meta.client
    metrics_data = _build_cloudwatch_metrics(
        metrics_name, metrics_type, value, dimension_types, ct, **kwargs
    )
    job_run_id = metrics_config.job_run_id
    try:
        response = cloudwatch_client.put_metric_data(
            Namespace=f"{DEFAULT_DELTACAT_METRICS_NAMESPACE}_{job_run_id}",
            MetricData=metrics_data,
        )
    except Exception as e:
        logger.warning(
            f"Failed to publish Cloudwatch metrics with name: {metrics_name}, "
            f"type: {metrics_type}, with exception: {e}, response: {response}"
        )


def emit_timer_metrics(metrics_name, value, metrics_config, **kwargs):
    metrics_dimension_type = [
        MetricsDimensionType.NODE_IP,
        MetricsDimensionType.RAY_TASK_ID,
        MetricsDimensionType.RAY_WORKER_ID,
    ]
    _emit_metrics(
        metrics_name=metrics_name,
        metrics_type=MetricsType.TIMER,
        metrics_config=metrics_config,
        value=value,
        dimension_types=metrics_dimension_type,
        **kwargs,
    )



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations
from enum import Enum
from typing import List
import datetime as dt
from datetime import timezone

TEST_S3_RCF_BUCKET_NAME = "test-compaction-artifacts-bucket"
# REBASE  src = spark compacted table to create an initial version of ray compacted table
BASE_TEST_SOURCE_NAMESPACE = "source_test_namespace"
BASE_TEST_SOURCE_TABLE_NAME = "test_table"
BASE_TEST_SOURCE_TABLE_VERSION = "1"

BASE_TEST_DESTINATION_NAMESPACE = "destination_test_namespace"
BASE_TEST_DESTINATION_TABLE_NAME = "destination_test_table_RAY"
BASE_TEST_DESTINATION_TABLE_VERSION = "1"

HASH_BUCKET_COUNT: int = 3

MAX_RECORDS_PER_FILE: int = 1

UTC_ISO_8601_FORMAT_WITHOUT_MILLIS = "%Y-%m-%dT%H:%M:%SZ"  # '2018-09-05T14:09:03Z'


class PartitionKeyType(str, Enum):
    INT = "int"
    STRING = "string"
    TIMESTAMP = "timestamp"


class PartitionKey(dict):
    @staticmethod
    def of(key_name: str, key_type: PartitionKeyType) -> PartitionKey:
        return PartitionKey({"keyName": key_name, "keyType": key_type.value})

    @property
    def key_name(self) -> str:
        return self["keyName"]

    @property
    def key_type(self) -> PartitionKeyType:
        key_type = self["keyType"]
        return None if key_type is None else PartitionKeyType(key_type)


def setup_sort_and_partition_keys(sort_keys_param, partition_keys_param):
    from deltacat.storage.model.sort_key import SortKey

    sort_keys, partition_keys = None, None
    if sort_keys_param is not None:
        sort_keys = [SortKey.of(sort_key["key_name"]) for sort_key in sort_keys_param]
    if partition_keys_param is not None:
        partition_keys = [
            PartitionKey.of(
                partition_key["key_name"], PartitionKeyType(partition_key["key_type"])
            )
            for partition_key in partition_keys_param
        ]
    return sort_keys, partition_keys


def offer_iso8601_timestamp_list(
    periods: int,
    unit_of_time: str,
    end_time=dt.datetime(2023, 5, 3, 10, 0, 0, 0, tzinfo=timezone.utc),
) -> List[str]:
    """
    Returns a list of ISO 8601 timestamps, each periods units of time before the start time.

    Args:
    periods: The number of timestamps to return.
    unit_of_time: The unit of time to use for the timestamps. Must be one of "seconds", "minutes", "hours", "days", or "weeks".
    end_time: The end time for the timestamps. Defaults to 2023-05-03T10:00:00Z.

    Returns:
    A list of ISO 8601 timestamps, each periods units of time before the start time.

    Raises:
    ValueError: If the unit_of_time argument is not one of "seconds", "minutes", "hours", "days", or "weeks".
    """
    import datetime as dt

    acceptable_units_of_time = ["seconds", "minutes", "hours", "days", "weeks"]
    if unit_of_time not in acceptable_units_of_time:
        raise ValueError(
            f"unit_of_time {unit_of_time} is not supported. Please use one of these time units: {acceptable_units_of_time}"
        )
    res = []
    for i in range(periods):
        kwarg = {unit_of_time: i}
        res.append(
            (end_time - dt.timedelta(**kwarg)).strftime(
                UTC_ISO_8601_FORMAT_WITHOUT_MILLIS
            )
        )
    return res



import logging
from ray import cloudpickle
from collections import defaultdict
import time
from deltacat.io.object_store import IObjectStore
from typing import Any, List, Optional
from deltacat import logs
import uuid
import socket
from pymemcache.client.base import Client
from pymemcache.client.retrying import RetryingClient
from pymemcache.exceptions import MemcacheUnexpectedCloseError
from pymemcache.client.rendezvous import RendezvousHash

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class MemcachedObjectStore(IObjectStore):
    """
    An implementation of object store that uses Memcached.
    """

    CONNECT_TIMEOUT = 10 * 60
    FETCH_TIMEOUT = 30 * 60

    def __init__(
        self,
        storage_node_ips: Optional[List[str]] = None,
        port: Optional[int] = 11212,
        connect_timeout: float = CONNECT_TIMEOUT,
        timeout: float = FETCH_TIMEOUT,
    ) -> None:
        self.client_cache = {}
        self.current_ip = None
        self.SEPARATOR = "_"
        self.port = port
        self.storage_node_ips = storage_node_ips
        self.hasher = None
        self.connect_timeout = connect_timeout
        self.timeout = timeout
        logger.info(f"The storage node IPs: {self.storage_node_ips}")
        super().__init__()

    def initialize_hasher(self):
        if not self.hasher and self.storage_node_ips:
            self.hasher = RendezvousHash()
            for n in self.storage_node_ips:
                self.hasher.add_node(n)

    def put_many(self, objects: List[object], *args, **kwargs) -> List[Any]:
        input = defaultdict(dict)
        result = []
        for obj in objects:
            serialized = cloudpickle.dumps(obj)
            uid = uuid.uuid4()
            create_ref_ip = self._get_create_ref_ip(uid.__str__())
            ref = self._create_ref(uid, create_ref_ip)
            input[create_ref_ip][uid.__str__()] = serialized
            result.append(ref)
        for create_ref_ip, uid_to_object in input.items():
            client = self._get_client_by_ip(create_ref_ip)
            if client.set_many(uid_to_object, noreply=False):
                raise RuntimeError("Unable to write few keys to cache")

        return result

    def put(self, obj: object, *args, **kwargs) -> Any:
        serialized = cloudpickle.dumps(obj)
        uid = uuid.uuid4()
        create_ref_ip = self._get_create_ref_ip(uid.__str__())
        ref = self._create_ref(uid, create_ref_ip)
        client = self._get_client_by_ip(create_ref_ip)

        if client.set(uid.__str__(), serialized):
            return ref
        else:
            raise RuntimeError("Unable to write to cache")

    def get_many(self, refs: List[Any], *args, **kwargs) -> List[object]:
        result = []
        uid_per_ip = defaultdict(lambda: [])

        start = time.monotonic()
        for ref in refs:
            uid, ip = ref.split(self.SEPARATOR)
            uid_per_ip[ip].append(uid)

        for (ip, uids) in uid_per_ip.items():
            client = self._get_client_by_ip(ip)
            cache_result = client.get_many(uids)
            assert len(cache_result) == len(
                uids
            ), f"Not all values were returned from cache as {len(cache_result)} != {len(uids)}"

            values = cache_result.values()
            total_bytes = 0

            deserialize_start = time.monotonic()
            for serialized in values:
                deserialized = cloudpickle.loads(serialized)
                total_bytes += len(serialized)
                result.append(deserialized)

            deserialize_end = time.monotonic()
            logger.debug(
                f"The time taken to deserialize {total_bytes} bytes is: {deserialize_end - deserialize_start}",
            )

        end = time.monotonic()

        logger.info(f"The total time taken to read all objects is: {end - start}")
        return result

    def get(self, ref: Any, *args, **kwargs) -> object:
        uid, ip = ref.split(self.SEPARATOR)
        client = self._get_client_by_ip(ip)
        serialized = client.get(uid)
        return cloudpickle.loads(serialized)

    def _create_ref(self, uid, ip) -> str:
        return f"{uid}{self.SEPARATOR}{ip}"

    def _get_storage_node_ip(self, key: str):
        self.initialize_hasher()
        storage_node_ip = self.hasher.get_node(key)
        return storage_node_ip

    def _get_create_ref_ip(self, uid: str):
        if self.storage_node_ips:
            create_ref_ip = self._get_storage_node_ip(uid)
        else:
            create_ref_ip = self._get_current_ip()
        return create_ref_ip

    def _get_client_by_ip(self, ip_address: str):
        if ip_address in self.client_cache:
            return self.client_cache[ip_address]

        base_client = Client(
            (ip_address, self.port),
            connect_timeout=self.connect_timeout,
            timeout=self.timeout,
            no_delay=True,
        )

        client = RetryingClient(
            base_client,
            attempts=15,
            retry_delay=1,
            retry_for=[
                MemcacheUnexpectedCloseError,
                ConnectionResetError,
                BrokenPipeError,
                TimeoutError,
            ],
        )

        self.client_cache[ip_address] = client
        return client

    def _get_current_ip(self):
        if self.current_ip is None:
            self.current_ip = socket.gethostbyname(socket.gethostname())

        return self.current_ip



from unittest.mock import patch
import unittest
from http import HTTPStatus
import requests

HAPPY_RESPONSE = {
    "AccessKeyId": "ASIA123456789",
    "Code": "Success",
    "Expiration": "2023-08-02T13:50:33Z",
    "LastUpdated": "2023-08-02T07:23:35Z",
    "SecretAccessKey": "bar",
    "Token": "foo",
    "Type": "AWS-HMAC",
}


class MockResponse:
    """
    A mock object denoting the response of requests method.
    """

    def __init__(self, status_code: int, text: str, reason: str = "") -> None:
        self.status_code: requests.Response.status_code = status_code
        self.text = text
        self.reason = reason

    def raise_for_status(*args, **kwargs):
        pass


class TestBlockUntilInstanceMetadataServiceReturnsSuccess(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        pass

    @patch("deltacat.aws.clients.requests")
    def test_sanity(self, requests_mock):
        from deltacat.aws.clients import (
            block_until_instance_metadata_service_returns_success,
        )

        requests_mock.get.return_value = MockResponse(200, "foo")
        self.assertEqual(
            block_until_instance_metadata_service_returns_success().status_code, 200
        )

    @patch("deltacat.aws.clients.requests")
    def test_retrying_on_statuses_in_status_force_list(self, requests_mock):
        from deltacat.aws.clients import (
            block_until_instance_metadata_service_returns_success,
        )

        requests_mock.get.side_effect = [
            MockResponse(HTTPStatus.OK, "foo"),
            MockResponse(HTTPStatus.TOO_MANY_REQUESTS, "foo"),
            MockResponse(HTTPStatus.INTERNAL_SERVER_ERROR, "foo"),
            MockResponse(HTTPStatus.NOT_IMPLEMENTED, "bar"),
            MockResponse(HTTPStatus.SERVICE_UNAVAILABLE, "bar"),
            MockResponse(HTTPStatus.GATEWAY_TIMEOUT, "bar"),
        ]
        self.assertEqual(
            block_until_instance_metadata_service_returns_success().status_code, 200
        )

    @patch("deltacat.aws.clients.requests")
    def test_retrying_status_on_shortlist_returns_early(self, requests_mock):
        from deltacat.aws.clients import (
            block_until_instance_metadata_service_returns_success,
        )

        requests_mock.get.side_effect = [
            MockResponse(HTTPStatus.FORBIDDEN, "foo"),
        ]
        self.assertEqual(
            block_until_instance_metadata_service_returns_success().status_code, 403
        )



import ray
from moto import mock_s3
import pytest
import os
import json
import boto3
from typing import Any, Dict, List, Optional, Set
from boto3.resources.base import ServiceResource
import pyarrow as pa
from deltacat.tests.test_utils.utils import read_s3_contents
from deltacat.tests.compute.common import (
    setup_sort_and_partition_keys,
    PartitionKey,
    TEST_S3_RCF_BUCKET_NAME,
    BASE_TEST_SOURCE_NAMESPACE,
    BASE_TEST_SOURCE_TABLE_NAME,
    BASE_TEST_DESTINATION_NAMESPACE,
    BASE_TEST_DESTINATION_TABLE_NAME,
)
from deltacat.tests.compute.testcases import (
    INCREMENTAL_TEST_CASES,
)

DATABASE_FILE_PATH_KEY, DATABASE_FILE_PATH_VALUE = (
    "db_file_path",
    "deltacat/tests/local_deltacat_storage/db_test.sqlite",
)

"""
MODULE scoped fixtures
"""


@pytest.fixture(autouse=True, scope="module")
def mock_aws_credential():
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_ID"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
    yield


@pytest.fixture(autouse=True, scope="module")
def cleanup_the_database_file_after_all_compaction_session_package_tests_complete():
    # make sure the database file is deleted after all the compactor package tests are completed
    if os.path.exists(DATABASE_FILE_PATH_VALUE):
        os.remove(DATABASE_FILE_PATH_VALUE)


@pytest.fixture(scope="module")
def setup_s3_resource(mock_aws_credential):
    with mock_s3():
        yield boto3.resource("s3")


@pytest.fixture(scope="module")
def setup_compaction_artifacts_s3_bucket(setup_s3_resource: ServiceResource):
    setup_s3_resource.create_bucket(
        ACL="authenticated-read",
        Bucket=TEST_S3_RCF_BUCKET_NAME,
    )
    yield


"""
FUNCTION scoped fixtures
"""


@pytest.fixture(scope="function")
def teardown_local_deltacat_storage_db():
    if os.path.exists(DATABASE_FILE_PATH_VALUE):
        os.remove(DATABASE_FILE_PATH_VALUE)


@pytest.fixture(scope="function")
def setup_local_deltacat_storage_conn():
    # see deltacat/tests/local_deltacat_storage/README.md for documentation
    kwargs_for_local_deltacat_storage: Dict[str, Any] = {
        DATABASE_FILE_PATH_KEY: DATABASE_FILE_PATH_VALUE,
    }
    yield kwargs_for_local_deltacat_storage


def setup_incremental_source_and_destination_tables(
    source_table_version: str,
    destination_table_version: str,
    primary_keys: Set[str],
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    column_names: List[str],
    arrow_arrays: List[pa.Array],
    partition_values: Optional[List[Any]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
    source_namespace: str = BASE_TEST_SOURCE_NAMESPACE,
    source_table_name: str = BASE_TEST_SOURCE_TABLE_NAME,
    destination_namespace: str = BASE_TEST_DESTINATION_NAMESPACE,
    destination_table_name: str = BASE_TEST_DESTINATION_TABLE_NAME,
):
    import deltacat.tests.local_deltacat_storage as ds
    from deltacat.types.media import ContentType
    from deltacat.storage import Partition, Stream

    ds.create_namespace(source_namespace, {}, **ds_mock_kwargs)
    ds.create_table_version(
        source_namespace,
        source_table_name,
        source_table_version,
        primary_key_column_names=list(primary_keys),
        sort_keys=sort_keys,
        partition_keys=partition_keys,
        supported_content_types=[ContentType.PARQUET],
        **ds_mock_kwargs,
    )
    source_table_stream: Stream = ds.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )
    test_table: pa.Table = pa.Table.from_arrays(arrow_arrays, names=column_names)
    staged_partition: Partition = ds.stage_partition(
        source_table_stream, partition_values, **ds_mock_kwargs
    )
    ds.commit_delta(
        ds.stage_delta(test_table, staged_partition, **ds_mock_kwargs), **ds_mock_kwargs
    )
    ds.commit_partition(staged_partition, **ds_mock_kwargs)
    # create the destination table
    ds.create_namespace(destination_namespace, {}, **ds_mock_kwargs)
    ds.create_table_version(
        destination_namespace,
        destination_table_name,
        destination_table_version,
        primary_key_column_names=list(primary_keys),
        sort_keys=sort_keys,
        partition_keys=partition_keys,
        supported_content_types=[ContentType.PARQUET],
        **ds_mock_kwargs,
    )
    destination_table_stream: Stream = ds.get_stream(
        namespace=destination_namespace,
        table_name=destination_table_name,
        table_version=destination_table_version,
        **ds_mock_kwargs,
    )
    source_table_stream_after_committed: Stream = ds.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )
    return source_table_stream_after_committed, destination_table_stream


@pytest.mark.parametrize(
    [
        "test_name",
        "source_table_version",
        "destination_table_version",
        "primary_keys_param",
        "sort_keys_param",
        "partition_keys_param",
        "column_names_param",
        "arrow_arrays_param",
        "rebase_source_partition_locator_param",
        "partition_values_param",
        "expected_result",
        "validation_callback_func",
        "validation_callback_func_kwargs",
        "do_teardown_local_deltacat_storage_db",
        "use_prev_compacted",
        "create_placement_group_param",
        "records_per_compacted_file_param",
        "hash_bucket_count_param",
        "compact_partition_func",
    ],
    [
        (
            test_name,
            source_table_version,
            destination_table_version,
            primary_keys_param,
            sort_keys_param,
            partition_keys_param,
            column_names_param,
            arrow_arrays_param,
            rebase_source_partition_locator_param,
            partition_values_param,
            expected_result,
            validation_callback_func,
            validation_callback_func_kwargs,
            do_teardown_local_deltacat_storage_db,
            use_prev_compacted,
            create_placement_group_param,
            records_per_compacted_file_param,
            hash_bucket_count_param,
            compact_partition_func,
        )
        for test_name, (
            source_table_version,
            destination_table_version,
            primary_keys_param,
            sort_keys_param,
            partition_keys_param,
            column_names_param,
            arrow_arrays_param,
            rebase_source_partition_locator_param,
            partition_values_param,
            expected_result,
            validation_callback_func,
            validation_callback_func_kwargs,
            do_teardown_local_deltacat_storage_db,
            use_prev_compacted,
            create_placement_group_param,
            records_per_compacted_file_param,
            hash_bucket_count_param,
            compact_partition_func,
        ) in INCREMENTAL_TEST_CASES.items()
    ],
    ids=[test_name for test_name in INCREMENTAL_TEST_CASES],
    indirect=[],
)
def test_compact_partition_incremental(
    request: pytest.FixtureRequest,
    setup_s3_resource: ServiceResource,
    setup_local_deltacat_storage_conn: Dict[str, Any],
    setup_compaction_artifacts_s3_bucket: None,
    test_name: str,
    source_table_version: str,
    destination_table_version: str,
    primary_keys_param: Set[str],
    sort_keys_param,
    partition_keys_param,
    column_names_param: List[str],
    arrow_arrays_param: List[pa.Array],
    rebase_source_partition_locator_param,
    partition_values_param,
    expected_result,
    # use and implement func and func_kwargs if you want to run additional validations apart from the ones in the test
    validation_callback_func,
    validation_callback_func_kwargs,
    do_teardown_local_deltacat_storage_db,
    use_prev_compacted,
    create_placement_group_param,
    records_per_compacted_file_param,
    hash_bucket_count_param,
    compact_partition_func,
):
    import deltacat.tests.local_deltacat_storage as ds
    from deltacat.types.media import ContentType
    from deltacat.storage import (
        PartitionLocator,
    )
    from deltacat.compute.compactor.model.compact_partition_params import (
        CompactPartitionParams,
    )
    from deltacat.utils.placement import (
        PlacementGroupManager,
    )
    from deltacat.compute.compactor import (
        RoundCompletionInfo,
    )

    ds_mock_kwargs = setup_local_deltacat_storage_conn

    # setup
    sort_keys, partition_keys = setup_sort_and_partition_keys(
        sort_keys_param, partition_keys_param
    )
    (
        source_table_stream,
        destination_table_stream,
    ) = setup_incremental_source_and_destination_tables(
        source_table_version,
        destination_table_version,
        primary_keys_param,
        sort_keys,
        partition_keys,
        column_names_param,
        arrow_arrays_param,
        partition_values_param,
        ds_mock_kwargs,
        f"{test_name}_src_namespace",
        f"{test_name}_table_src",
        f"{test_name}_dest_namespace",
        f"{test_name}_table_dest",
    )
    ray.shutdown()
    ray.init(local_mode=True)
    assert ray.is_initialized()
    source_partition = ds.get_partition(
        source_table_stream.locator,
        partition_values_param,
        **ds_mock_kwargs,
    )
    destination_partition_locator = PartitionLocator.of(
        destination_table_stream.locator,
        partition_values_param,
        None,
    )
    num_workers, worker_instance_cpu = 1, 1
    total_cpus = num_workers * worker_instance_cpu
    pgm = None
    if create_placement_group_param:
        pgm = PlacementGroupManager(
            1, total_cpus, worker_instance_cpu, memory_per_bundle=4000000
        ).pgs[0]
    compact_partition_params = CompactPartitionParams.of(
        {
            "compaction_artifact_s3_bucket": TEST_S3_RCF_BUCKET_NAME,
            "compacted_file_content_type": ContentType.PARQUET,
            "dd_max_parallelism_ratio": 1.0,
            "deltacat_storage": ds,
            "deltacat_storage_kwargs": ds_mock_kwargs,
            "destination_partition_locator": destination_partition_locator,
            "hash_bucket_count": hash_bucket_count_param,
            "last_stream_position_to_compact": source_partition.stream_position,
            "list_deltas_kwargs": {**ds_mock_kwargs, **{"equivalent_table_types": []}},
            "pg_config": pgm,
            "primary_keys": primary_keys_param,
            "rebase_source_partition_locator": rebase_source_partition_locator_param,
            "records_per_compacted_file": records_per_compacted_file_param,
            "s3_client_kwargs": {},
            "source_partition_locator": source_partition.locator,
            "sort_keys": sort_keys if sort_keys else None,
        }
    )
    # execute
    rcf_file_s3_uri = compact_partition_func(compact_partition_params)
    # validate
    _, rcf_object_key = rcf_file_s3_uri.rsplit("/", 1)
    rcf_file_output: Dict[str, Any] = read_s3_contents(
        setup_s3_resource, TEST_S3_RCF_BUCKET_NAME, rcf_object_key
    )
    round_completion_info = RoundCompletionInfo(**rcf_file_output)
    print(f"rcf_file_output: {json.dumps(rcf_file_output, indent=2)}")
    compacted_delta_locator = round_completion_info.compacted_delta_locator
    tables = ds.download_delta(compacted_delta_locator, **ds_mock_kwargs)
    compacted_table = pa.concat_tables(tables)

    # the compacted table may contain multiple files and chunks
    # and order of records may be incorrect due to multiple files.
    expected_result = expected_result.combine_chunks().sort_by(
        [(val, "ascending") for val in primary_keys_param]
    )
    compacted_table = compacted_table.combine_chunks().sort_by(
        [(val, "ascending") for val in primary_keys_param]
    )

    assert compacted_table.equals(
        expected_result
    ), f"{compacted_table} does not match {expected_result}"
    if (
        validation_callback_func is not None
        and validation_callback_func_kwargs is not None
    ):
        validation_callback_func(**validation_callback_func_kwargs)
    # https://docs.pytest.org/en/7.1.x/reference/reference.html#pytest.FixtureRequest.getfixturevalue
    if do_teardown_local_deltacat_storage_db:
        request.getfixturevalue("teardown_local_deltacat_storage_db")
    return



import pyarrow as pa
from typing import Dict, List
from deltacat.tests.compute.common import (
    MAX_RECORDS_PER_FILE,
    offer_iso8601_timestamp_list,
)
from deltacat.tests.compute.common import (
    BASE_TEST_SOURCE_TABLE_VERSION,
    BASE_TEST_DESTINATION_TABLE_VERSION,
    HASH_BUCKET_COUNT,
)
from deltacat.compute.compactor.compaction_session import (
    compact_partition_from_request as compact_partition_v1,
)
from deltacat.compute.compactor_v2.compaction_session import (
    compact_partition as compact_partition_v2,
)


def create_tests_cases_for_all_compactor_versions(test_cases: Dict[str, List]):
    final_cases = {}
    for version, compact_partition_func in enumerate(
        [compact_partition_v1, compact_partition_v2]
    ):
        for case_name, case_value in test_cases.items():
            final_cases[f"{case_name}_v{version + 1}"] = [
                *case_value,
                compact_partition_func,
            ]

    return final_cases


"""
TODO Test Cases:
1. incremental w/wout round completion file
2. Backfill w/wout round completion
3. Rebase w/wout round completion file
4. Rebase then incremental (use same round completion file)
"""


INCREMENTAL_INDEPENDENT_TEST_CASES = {
    "1-incremental-pkstr-sknone-norcf": [
        BASE_TEST_SOURCE_TABLE_VERSION,
        BASE_TEST_DESTINATION_TABLE_VERSION,
        {"pk_col_1"},  # Primary key columns
        [],  # Sort key columns
        [{"key_name": "region_id", "key_type": "int"}],  # Partition keys
        ["pk_col_1"],  # column_names
        [pa.array([str(i) for i in range(10)])],  # arrow arrays
        None,  # rebase_source_partition_locator_param
        ["1"],  # partition_values_param
        pa.Table.from_arrays(
            [pa.array([str(i) for i in range(10)])],
            names=["pk_col_1"],  # expected_result
        ),
        None,  # validation_callback_func
        None,  # validation_callback_func_kwargs
        True,  # teardown_local_deltacat_storage_db
        False,  # use_prev_compacted
        True,  # create_placement_group_param
        MAX_RECORDS_PER_FILE,  # records_per_compacted_file_param
        HASH_BUCKET_COUNT,  # hash_bucket_count_param
    ],
    "2-incremental-pkstr-skstr-norcf": [
        BASE_TEST_SOURCE_TABLE_VERSION,
        BASE_TEST_DESTINATION_TABLE_VERSION,
        ["pk_col_1"],
        [
            {
                "key_name": "sk_col_1",
            }
        ],
        [],
        ["pk_col_1", "sk_col_1"],
        [pa.array([str(i) for i in range(10)]), pa.array(["test"] * 10)],
        None,
        ["1"],
        pa.Table.from_arrays(
            [pa.array([str(i) for i in range(10)]), pa.array(["test"] * 10)],
            names=["pk_col_1", "sk_col_1"],
        ),
        None,
        None,
        True,
        False,
        True,
        MAX_RECORDS_PER_FILE,
        HASH_BUCKET_COUNT,
    ],
    "3-incremental-pkstr-multiskstr-norcf": [
        BASE_TEST_SOURCE_TABLE_VERSION,
        BASE_TEST_DESTINATION_TABLE_VERSION,
        ["pk_col_1"],
        [
            {
                "key_name": "sk_col_1",
            },
            {
                "key_name": "sk_col_2",
            },
        ],
        [],
        ["pk_col_1", "sk_col_1", "sk_col_2"],
        [
            pa.array([str(i) for i in range(10)]),
            pa.array(["test"] * 10),
            pa.array(["foo"] * 10),
        ],
        None,
        ["1"],
        pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array(["test"] * 10),
                pa.array(["foo"] * 10),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2"],
        ),
        None,
        None,
        True,
        False,
        True,
        MAX_RECORDS_PER_FILE,
        HASH_BUCKET_COUNT,
    ],
    "4-incremental-duplicate-pk": [
        BASE_TEST_SOURCE_TABLE_VERSION,
        BASE_TEST_DESTINATION_TABLE_VERSION,
        ["pk_col_1"],
        [
            {
                "key_name": "sk_col_1",
            },
            {
                "key_name": "sk_col_2",
            },
        ],
        [],
        ["pk_col_1", "sk_col_1", "sk_col_2"],
        [
            pa.array([str(i) for i in range(5)] + ["6", "6", "6", "6", "6"]),
            pa.array([str(i) for i in range(10)]),
            pa.array(["foo"] * 10),
        ],
        None,
        ["1"],
        pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(5)] + ["6"]),
                pa.array([str(i) for i in range(5)] + ["9"]),
                pa.array(["foo"] * 6),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2"],
        ),
        None,
        None,
        True,
        False,
        True,
        MAX_RECORDS_PER_FILE,
        HASH_BUCKET_COUNT,
    ],
    "5-incremental-decimal-pk-simple": [
        BASE_TEST_SOURCE_TABLE_VERSION,
        BASE_TEST_DESTINATION_TABLE_VERSION,
        ["pk_col_1"],
        [
            {
                "key_name": "sk_col_1",
            },
        ],
        [],
        ["pk_col_1", "sk_col_1"],
        [
            pa.array([i / 10 for i in range(0, 10)]),
            pa.array([str(i) for i in range(10)]),
        ],
        None,
        ["1"],
        pa.Table.from_arrays(
            [
                pa.array([i / 10 for i in range(0, 10)]),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        None,
        None,
        True,
        False,
        True,
        MAX_RECORDS_PER_FILE,
        HASH_BUCKET_COUNT,
    ],
    "7-incremental-integer-pk-simple": [
        BASE_TEST_SOURCE_TABLE_VERSION,
        BASE_TEST_DESTINATION_TABLE_VERSION,
        ["pk_col_1"],
        [
            {
                "key_name": "sk_col_1",
            },
        ],
        [],
        ["pk_col_1", "sk_col_1"],
        [
            pa.array([i for i in range(0, 10)]),
            pa.array([str(i) for i in range(10)]),
        ],
        None,
        ["1"],
        pa.Table.from_arrays(
            [
                pa.array([i for i in range(0, 10)]),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        None,
        None,
        True,
        False,
        True,
        MAX_RECORDS_PER_FILE,
        HASH_BUCKET_COUNT,
    ],
    "8-incremental-timestamp-pk-simple": [
        BASE_TEST_SOURCE_TABLE_VERSION,
        BASE_TEST_DESTINATION_TABLE_VERSION,
        ["pk_col_1"],
        [
            {
                "key_name": "sk_col_1",
            },
        ],
        [],
        ["pk_col_1", "sk_col_1"],
        [
            pa.array(offer_iso8601_timestamp_list(10, "minutes")),
            pa.array([str(i) for i in range(10)]),
        ],
        None,
        ["1"],
        pa.Table.from_arrays(
            [
                pa.array(offer_iso8601_timestamp_list(10, "minutes")),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        None,
        None,
        True,
        False,
        True,
        MAX_RECORDS_PER_FILE,
        HASH_BUCKET_COUNT,
    ],
    "9-incremental-decimal-timestamp-pk-multi": [
        BASE_TEST_SOURCE_TABLE_VERSION,
        BASE_TEST_DESTINATION_TABLE_VERSION,
        ["pk_col_1", "pk_col_2"],
        [
            {
                "key_name": "sk_col_1",
            },
        ],
        [],
        ["pk_col_1", "pk_col_2", "sk_col_1"],
        [
            pa.array([i / 10 for i in range(0, 20)]),
            pa.array(offer_iso8601_timestamp_list(20, "minutes")),
            pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
        ],
        None,
        ["1"],
        pa.Table.from_arrays(
            [
                pa.array([i / 10 for i in range(0, 20)]),
                pa.array(offer_iso8601_timestamp_list(20, "minutes")),
                pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
            ],
            names=["pk_col_1", "pk_col_2", "sk_col_1"],
        ),
        None,
        None,
        True,
        False,
        True,
        MAX_RECORDS_PER_FILE,
        HASH_BUCKET_COUNT,
    ],
    "10-incremental-decimal-pk-multi-dup": [
        BASE_TEST_SOURCE_TABLE_VERSION,
        BASE_TEST_DESTINATION_TABLE_VERSION,
        ["pk_col_1"],
        [
            {
                "key_name": "sk_col_1",
            },
        ],
        [],
        ["pk_col_1", "sk_col_1"],
        [
            pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
            pa.array(reversed([i for i in range(20)])),
        ],
        None,
        ["1"],
        pa.Table.from_arrays(
            [
                pa.array([0.1, 0.2, 0.3, 0.4, 0.5]),
                pa.array([19, 15, 11, 7, 3]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        None,
        None,
        True,
        False,
        True,
        MAX_RECORDS_PER_FILE,
        HASH_BUCKET_COUNT,
    ],
}

"""
for test_name, (
            source_table_version,
            destination_table_version,
            primary_keys_param,
            sort_keys_param,
            partition_keys_param,
            column_names_param,
            arrow_arrays_param,
            rebase_source_partition_locator_param,
            partition_values_param,
            expected_result,
            validation_callback_func,
            validation_callback_func_kwargs,
            do_teardown_local_deltacat_storage_db,
            use_prev_compacted,
            create_placement_group_param,
            records_per_compacted_file_param,
            hash_bucket_count_param,
        ) in INCREMENTAL_TEST_CASES.items()
"""

# TODO: Add test cases where next tc is dependent on the previous compacted table existing
INCREMENTAL_DEPENDENT_TEST_CASES = {
    "11-incremental-multi-dup-retain-table": (
        BASE_TEST_SOURCE_TABLE_VERSION,
        BASE_TEST_DESTINATION_TABLE_VERSION,
        ["pk_col_1", "pk_col_2"],
        [
            {
                "key_name": "sk_col_1",
            },
        ],
        [],
        ["pk_col_1", "pk_col_2", "sk_col_1"],
        [
            pa.array([i / 10 for i in range(0, 20)]),
            pa.array(offer_iso8601_timestamp_list(20, "minutes")),
            pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
        ],
        None,
        ["1"],
        pa.Table.from_arrays(
            [
                pa.array([i / 10 for i in range(0, 20)]),
                pa.array(offer_iso8601_timestamp_list(20, "minutes")),
                pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
            ],
            names=["pk_col_1", "pk_col_2", "sk_col_1"],
        ),
        None,
        None,
        False,
        False,
        True,
        MAX_RECORDS_PER_FILE,
        HASH_BUCKET_COUNT,
    ),
}

INCREMENTAL_TEST_CASES = create_tests_cases_for_all_compactor_versions(
    {
        **INCREMENTAL_INDEPENDENT_TEST_CASES,
        **INCREMENTAL_DEPENDENT_TEST_CASES,
    }
)



from typing import Any, Callable, Dict, List, Optional, Set, Union, Tuple

import pyarrow as pa
import json
import sqlite3
from sqlite3 import Cursor, Connection
import uuid
import io
from deltacat.utils.common import current_time_ms

from deltacat.storage import (
    Delta,
    DeltaLocator,
    DeltaType,
    DistributedDataset,
    LifecycleState,
    ListResult,
    LocalDataset,
    LocalTable,
    Manifest,
    ManifestAuthor,
    Namespace,
    NamespaceLocator,
    Partition,
    SchemaConsistencyType,
    Stream,
    StreamLocator,
    Table,
    TableVersion,
    TableVersionLocator,
    TableLocator,
    CommitState,
    SortKey,
    PartitionLocator,
    ManifestMeta,
    ManifestEntry,
    ManifestEntryList,
)
from deltacat.types.media import ContentType, StorageType, TableType, ContentEncoding
from deltacat.utils.common import ReadKwargsProvider

SQLITE_CUR_ARG = "sqlite3_cur"
SQLITE_CON_ARG = "sqlite3_con"
DB_FILE_PATH_ARG = "db_file_path"

STORAGE_TYPE = "SQLITE3"
STREAM_ID_PROPERTY = "stream_id"
CREATE_NAMESPACES_TABLE = (
    "CREATE TABLE IF NOT EXISTS namespaces(locator, value, PRIMARY KEY (locator))"
)
CREATE_TABLES_TABLE = (
    "CREATE TABLE IF NOT EXISTS tables(locator, namespace_locator, value, PRIMARY KEY (locator), "
    "FOREIGN KEY (namespace_locator) REFERENCES namespaces(locator))"
)
CREATE_TABLE_VERSIONS_TABLE = (
    "CREATE TABLE IF NOT EXISTS table_versions(locator, table_locator, value, PRIMARY KEY (locator), "
    "FOREIGN KEY (table_locator) REFERENCES tables(locator))"
)
CREATE_STREAMS_TABLE = (
    "CREATE TABLE IF NOT EXISTS streams(locator, table_version_locator, value, PRIMARY KEY(locator), "
    "FOREIGN KEY (table_version_locator) REFERENCES table_versions(locator))"
)
CREATE_PARTITIONS_TABLE = (
    "CREATE TABLE IF NOT EXISTS partitions(locator, stream_locator, value, PRIMARY KEY(locator), "
    "FOREIGN KEY (stream_locator) REFERENCES streams(locator))"
)
CREATE_DELTAS_TABLE = (
    "CREATE TABLE IF NOT EXISTS deltas(locator, partition_locator, value, PRIMARY KEY(locator), "
    "FOREIGN KEY (partition_locator) REFERENCES partitions(locator))"
)
CREATE_DATA_TABLE = "CREATE TABLE IF NOT EXISTS data(uri, value, PRIMARY KEY(uri))"


def _get_sqlite3_cursor_con(kwargs) -> Tuple[Cursor, Connection]:
    if SQLITE_CUR_ARG in kwargs and SQLITE_CON_ARG in kwargs:
        return kwargs[SQLITE_CUR_ARG], kwargs[SQLITE_CON_ARG]
    elif DB_FILE_PATH_ARG in kwargs:
        con = sqlite3.connect(kwargs[DB_FILE_PATH_ARG])
        cur = con.cursor()
        return cur, con

    raise ValueError(f"Invalid local db connection kwargs: {kwargs}")


def _get_manifest_entry_uri(manifest_entry_id: str) -> str:
    return f"cloudpickle://{manifest_entry_id}"


def list_namespaces(*args, **kwargs) -> ListResult[Namespace]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    res = cur.execute("SELECT * FROM namespaces")
    fetched = res.fetchall()
    result = []

    for item in fetched:
        result.append(Namespace(json.loads(item[1])))

    return ListResult.of(result, None, None)


def list_tables(namespace: str, *args, **kwargs) -> ListResult[Table]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    params = (NamespaceLocator.of(namespace).canonical_string(),)
    res = cur.execute("SELECT * FROM tables WHERE namespace_locator = ?", params)
    fetched = res.fetchall()
    result = []

    for item in fetched:
        result.append(Table(json.loads(item[2])))

    return ListResult.of(result, None, None)


def list_table_versions(
    namespace: str, table_name: str, *args, **kwargs
) -> ListResult[TableVersion]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    table_locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)

    res = cur.execute(
        "SELECT * FROM table_versions WHERE table_locator = ?",
        (table_locator.canonical_string(),),
    )
    fetched = res.fetchall()
    result = []

    for item in fetched:
        result.append(TableVersion(json.loads(item[2])))

    return ListResult.of(result, None, None)


def list_partitions(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> ListResult[Partition]:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    stream = get_stream(namespace, table_name, table_version, *args, **kwargs)

    res = cur.execute(
        "SELECT * FROM partitions WHERE stream_locator = ?",
        (stream.locator.canonical_string(),),
    )

    fetched = res.fetchall()
    result = []
    for item in fetched:
        partition = Partition(json.loads(item[2]))
        if partition.state == CommitState.COMMITTED:
            result.append(partition)

    return ListResult.of(result, None, None)


def list_stream_partitions(stream: Stream, *args, **kwargs) -> ListResult[Partition]:
    return list_partitions(
        stream.namespace, stream.table_name, stream.table_version, *args, **kwargs
    )


def list_deltas(
    namespace: str,
    table_name: str,
    partition_values: Optional[List[Any]] = None,
    table_version: Optional[str] = None,
    first_stream_position: Optional[int] = None,
    last_stream_position: Optional[int] = None,
    ascending_order: Optional[bool] = None,
    include_manifest: bool = False,
    *args,
    **kwargs,
) -> ListResult[Delta]:
    stream = get_stream(namespace, table_name, table_version, *args, **kwargs)
    if stream is None:
        return ListResult.of([], None, None)

    partition = get_partition(stream.locator, partition_values, *args, **kwargs)

    all_deltas = list_partition_deltas(
        partition,
        first_stream_position=first_stream_position,
        last_stream_position=last_stream_position,
        ascending_order=ascending_order,
        include_manifest=include_manifest,
        *args,
        **kwargs,
    ).all_items()

    result = []

    for delta in all_deltas:
        if (
            not first_stream_position or first_stream_position < delta.stream_position
        ) and (
            not last_stream_position or delta.stream_position <= last_stream_position
        ):
            result.append(delta)

        if not include_manifest:
            delta.manifest = None

    result.sort(reverse=(not ascending_order), key=lambda d: d.stream_position)
    return ListResult.of(result, None, None)


def list_partition_deltas(
    partition_like: Union[Partition, PartitionLocator],
    first_stream_position: Optional[int] = None,
    last_stream_position: Optional[int] = None,
    ascending_order: bool = False,
    include_manifest: bool = False,
    *args,
    **kwargs,
) -> ListResult[Delta]:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    if partition_like is None:
        return ListResult.of([], None, None)

    if first_stream_position is None:
        first_stream_position = 0

    if last_stream_position is None:
        last_stream_position = float("inf")

    assert isinstance(partition_like, Partition) or isinstance(
        partition_like, PartitionLocator
    ), f"Expected a Partition or PartitionLocator as an input argument but found {partition_like}"

    partition_locator = None
    if isinstance(partition_like, Partition):
        partition_locator = partition_like.locator
    else:
        partition_locator = partition_like

    res = cur.execute(
        "SELECT * FROM deltas WHERE partition_locator = ?",
        (partition_locator.canonical_string(),),
    )

    serialized_items = res.fetchall()

    if not serialized_items:
        return ListResult.of([], None, None)

    result = []
    for item in serialized_items:
        current_delta = Delta(json.loads(item[2]))
        if (
            first_stream_position
            <= current_delta.stream_position
            <= last_stream_position
        ):
            result.append(current_delta)

        if not include_manifest:
            current_delta.manifest = None

    result.sort(
        reverse=True if not ascending_order else False, key=lambda d: d.stream_position
    )
    return ListResult.of(result, None, None)


def get_delta(
    namespace: str,
    table_name: str,
    stream_position: int,
    partition_values: Optional[List[Any]] = None,
    table_version: Optional[str] = None,
    include_manifest: bool = False,
    *args,
    **kwargs,
) -> Optional[Delta]:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    stream = get_stream(namespace, table_name, table_version, *args, **kwargs)
    partition = get_partition(stream.locator, partition_values, *args, **kwargs)
    delta_locator = DeltaLocator.of(partition.locator, stream_position)

    res = cur.execute(
        "SELECT * FROM deltas WHERE locator = ?", (delta_locator.canonical_string(),)
    )

    serialized_delta = res.fetchone()
    if serialized_delta is None:
        return None

    delta = Delta(json.loads(serialized_delta[2]))

    if not include_manifest:
        delta.manifest = None

    return delta


def get_latest_delta(
    namespace: str,
    table_name: str,
    partition_values: Optional[List[Any]] = None,
    table_version: Optional[str] = None,
    include_manifest: bool = False,
    *args,
    **kwargs,
) -> Optional[Delta]:

    deltas = list_deltas(
        namespace,
        table_name,
        partition_values,
        table_version,
        None,
        None,
        False,
        include_manifest,
        *args,
        **kwargs,
    ).all_items()

    if not deltas:
        return None

    return deltas[0]


def download_delta(
    delta_like: Union[Delta, DeltaLocator],
    table_type: TableType = TableType.PYARROW,
    storage_type: StorageType = StorageType.DISTRIBUTED,
    max_parallelism: Optional[int] = None,
    columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    ray_options_provider: Callable[[int, Any], Dict[str, Any]] = None,
    *args,
    **kwargs,
) -> Union[LocalDataset, DistributedDataset]:
    result = []
    manifest = get_delta_manifest(delta_like, *args, **kwargs)

    for entry_index in range(len(manifest.entries)):
        result.append(
            download_delta_manifest_entry(
                delta_like=delta_like,
                entry_index=entry_index,
                table_type=table_type,
                columns=columns,
                file_reader_kwargs_provider=file_reader_kwargs_provider,
                *args,
                **kwargs,
            )
        )

    return result


def download_delta_manifest_entry(
    delta_like: Union[Delta, DeltaLocator],
    entry_index: int,
    table_type: TableType = TableType.PYARROW,
    columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    *args,
    **kwargs,
) -> LocalTable:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    manifest = get_delta_manifest(delta_like, *args, **kwargs)
    if entry_index >= len(manifest.entries):
        raise IndexError(
            f"Manifest entry index {entry_index} does not exist. "
            f"Valid values: [0, {len(manifest.entries)}]"
        )

    entry = manifest.entries[entry_index]

    res = cur.execute("SELECT value FROM data WHERE uri = ?", (entry.uri,))
    serialized_data = res.fetchone()

    if serialized_data is None:
        raise ValueError(
            f"Invalid value of delta locator: {delta_like.canonical_string()}"
        )

    serialized_data = serialized_data[0]
    if entry.meta.content_type == ContentType.PARQUET:
        if table_type == TableType.PYARROW_PARQUET:
            table = pa.parquet.ParquetFile(io.BytesIO(serialized_data))
        else:
            table = pa.parquet.read_table(io.BytesIO(serialized_data), columns=columns)
    elif entry.meta.content_type == ContentType.UNESCAPED_TSV:
        assert (
            table_type != TableType.PYARROW_PARQUET
        ), f"uTSV table cannot be read as {table_type}"
        parse_options = pa.csv.ParseOptions(delimiter="\t")
        convert_options = pa.csv.ConvertOptions(
            null_values=[""], strings_can_be_null=True, include_columns=columns
        )
        table = pa.csv.read_csv(
            io.BytesIO(serialized_data),
            parse_options=parse_options,
            convert_options=convert_options,
        )
    else:
        raise ValueError(f"Content type: {entry.meta.content_type} not supported.")

    if table_type == TableType.PYARROW:
        return table
    elif table_type == TableType.PYARROW_PARQUET:
        return table
    elif table_type == TableType.NUMPY:
        raise NotImplementedError(f"Table type={table_type} not supported")
    elif table_type == TableType.PANDAS:
        return table.to_pandas()

    return table


def get_delta_manifest(
    delta_like: Union[Delta, DeltaLocator], *args, **kwargs
) -> Optional[Manifest]:
    delta = get_delta(
        namespace=delta_like.namespace,
        table_name=delta_like.table_name,
        stream_position=delta_like.stream_position,
        partition_values=delta_like.partition_values,
        table_version=delta_like.table_version,
        include_manifest=True,
        *args,
        **kwargs,
    )
    if not delta:
        return None

    return delta.manifest


def create_namespace(
    namespace: str, permissions: Dict[str, Any], *args, **kwargs
) -> Namespace:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = NamespaceLocator.of(namespace)
    result = Namespace.of(locator, permissions)
    params = (locator.canonical_string(), json.dumps(result))
    cur.execute(CREATE_NAMESPACES_TABLE)
    cur.execute(CREATE_TABLES_TABLE)
    cur.execute(CREATE_TABLE_VERSIONS_TABLE)
    cur.execute(CREATE_STREAMS_TABLE)
    cur.execute(CREATE_PARTITIONS_TABLE)
    cur.execute(CREATE_DELTAS_TABLE)
    cur.execute(CREATE_DATA_TABLE)
    cur.execute("INSERT OR IGNORE INTO namespaces VALUES(?, ?)", params)
    con.commit()
    return result


def update_namespace(
    namespace: str,
    permissions: Optional[Dict[str, Any]] = None,
    new_namespace: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    assert new_namespace is None, "namespace name cannot be changed"
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = NamespaceLocator.of(namespace)
    result = Namespace.of(locator, permissions)
    params = (json.dumps(result), locator.canonical_string())
    cur.execute("UPDATE namespaces SET value = ? WHERE locator = ?", params)
    con.commit()


def create_table_version(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    schema: Optional[Union[pa.Schema, str, bytes]] = None,
    schema_consistency: Optional[Dict[str, SchemaConsistencyType]] = None,
    partition_keys: Optional[List[Dict[str, Any]]] = None,
    primary_key_column_names: Optional[Set[str]] = None,
    sort_keys: Optional[List[SortKey]] = None,
    table_version_description: Optional[str] = None,
    table_version_properties: Optional[Dict[str, str]] = None,
    table_permissions: Optional[Dict[str, Any]] = None,
    table_description: Optional[str] = None,
    table_properties: Optional[Dict[str, str]] = None,
    supported_content_types: Optional[List[ContentType]] = None,
    *args,
    **kwargs,
) -> Stream:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    latest_version = get_latest_table_version(namespace, table_name, *args, **kwargs)
    if (
        table_version is not None
        and latest_version
        and int(latest_version.table_version) + 1 != int(table_version)
    ):
        raise AssertionError(
            f"Table version can only be incremented. Last version={latest_version.table_version}"
        )
    elif table_version is None:
        table_version = (
            (int(latest_version.table_version) + 1) if latest_version else "1"
        )

    table_locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)
    table_obj = Table.of(
        table_locator, table_permissions, table_description, table_properties
    )
    table_version_locator = TableVersionLocator.of(
        table_locator=table_locator, table_version=table_version
    )

    stream_id = uuid.uuid4().__str__()

    if table_version_properties is None:
        table_version_properties = {}

    properties = {**table_version_properties, STREAM_ID_PROPERTY: stream_id}
    table_version_obj = TableVersion.of(
        table_version_locator,
        schema=schema,
        partition_keys=partition_keys,
        primary_key_columns=primary_key_column_names,
        description=table_version_description,
        properties=properties,
        sort_keys=sort_keys,
        content_types=supported_content_types,
    )
    stream_locator = StreamLocator.of(
        table_version_obj.locator, stream_id=stream_id, storage_type=STORAGE_TYPE
    )
    result_stream = Stream.of(
        stream_locator, partition_keys=partition_keys, state=CommitState.COMMITTED
    )

    params = (
        table_locator.canonical_string(),
        table_locator.namespace_locator.canonical_string(),
        json.dumps(table_obj),
    )
    cur.execute("INSERT OR IGNORE INTO tables VALUES (?, ?, ?)", params)
    params = (
        table_version_locator.canonical_string(),
        table_locator.canonical_string(),
        json.dumps(table_version_obj),
    )
    cur.execute("INSERT OR IGNORE INTO table_versions VALUES (?, ?, ?)", params)

    params = (
        stream_locator.canonical_string(),
        table_version_locator.canonical_string(),
        json.dumps(result_stream),
    )
    cur.execute("INSERT OR IGNORE INTO streams VALUES (?, ?, ?)", params)
    con.commit()
    return result_stream


def update_table(
    namespace: str,
    table_name: str,
    permissions: Optional[Dict[str, Any]] = None,
    description: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    new_table_name: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    table_locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)
    table_obj = Table.of(table_locator, permissions, description, properties)

    params = (table_locator.canonical_string(),)
    cur.execute("DELETE FROM tables WHERE locator = ?", params)
    params = (
        table_locator.canonical_string(),
        table_locator.namespace_locator.canonical_string(),
        json.dumps(table_obj),
    )
    cur.execute("INSERT INTO tables VALUES (?, ?, ?)", params)
    con.commit()


def update_table_version(
    namespace: str,
    table_name: str,
    table_version: str,
    lifecycle_state: Optional[LifecycleState] = None,
    schema: Optional[Union[pa.Schema, str, bytes]] = None,
    schema_consistency: Optional[Dict[str, SchemaConsistencyType]] = None,
    description: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    *args,
    **kwargs,
) -> None:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    table_locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)
    table_version_locator = TableVersionLocator.of(
        table_locator=table_locator, table_version=table_version
    )

    res = cur.execute(
        "SELECT * from table_versions WHERE locator = ?",
        (table_version_locator.canonical_string(),),
    )
    serialized_table_version = res.fetchone()
    assert (
        serialized_table_version is not None
    ), f"Table version not found with locator={table_version_locator.canonical_string()}"
    current_table_version_obj = TableVersion(json.loads(serialized_table_version[2]))

    if properties is None:
        properties = {}

    current_props = (
        current_table_version_obj.properties
        if current_table_version_obj.properties
        else {}
    )

    tv_properties = {**properties, **current_props}
    table_version_obj = TableVersion.of(
        table_version_locator,
        schema=schema,
        partition_keys=current_table_version_obj.partition_keys,
        primary_key_columns=current_table_version_obj.primary_keys,
        description=description,
        properties=tv_properties,
        sort_keys=current_table_version_obj.sort_keys,
        content_types=current_table_version_obj.content_types,
    )

    params = (
        table_locator.canonical_string(),
        json.dumps(table_version_obj),
        table_version_locator.canonical_string(),
    )
    cur.execute(
        "UPDATE table_versions SET table_locator = ?, value = ? WHERE locator = ?",
        params,
    )
    con.commit()


def stage_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Stream:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    existing_table_version = get_table_version(
        namespace, table_name, table_version, *args, **kwargs
    )
    existing_stream = get_stream(namespace, table_name, table_version, *args, **kwargs)

    stream_id = uuid.uuid4().__str__()
    new_stream_locator = StreamLocator.of(
        existing_table_version.locator, stream_id, STORAGE_TYPE
    )
    new_stream = Stream.of(
        new_stream_locator,
        existing_stream.partition_keys,
        CommitState.STAGED,
        existing_stream.locator.canonical_string(),
    )

    params = (
        new_stream_locator.canonical_string(),
        existing_table_version.locator.canonical_string(),
        json.dumps(new_stream),
    )
    cur.execute("INSERT INTO streams VALUES (?, ?, ?)", params)
    con.commit()

    return new_stream


def commit_stream(stream: Stream, *args, **kwargs) -> Stream:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    existing_table_version = get_table_version(
        stream.namespace, stream.table_name, stream.table_version, *args, **kwargs
    )
    stream_to_commit = Stream.of(
        stream.locator,
        stream.partition_keys,
        CommitState.COMMITTED,
        stream.previous_stream_digest,
    )

    existing_table_version.properties[
        STREAM_ID_PROPERTY
    ] = stream_to_commit.locator.stream_id

    params = (
        json.dumps(existing_table_version),
        existing_table_version.locator.canonical_string(),
    )
    cur.execute("UPDATE table_versions SET value = ? WHERE locator = ?", params)
    params = (json.dumps(stream_to_commit), stream_to_commit.locator.canonical_string())
    cur.execute("UPDATE streams SET value = ? WHERE locator = ?", params)
    con.commit()

    return stream_to_commit


def delete_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    table_version_locator = TableVersionLocator.of(
        TableLocator.of(NamespaceLocator.of(namespace), table_name), table_version
    )

    res = cur.execute(
        "SELECT locator FROM streams WHERE table_version_locator = ?",
        (table_version_locator.canonical_string(),),
    )
    locators = res.fetchall()
    cur.executemany("DELETE FROM streams WHERE locator = ?", locators)
    cur.execute(
        "DELETE FROM table_versions WHERE locator = ?",
        (table_version_locator.canonical_string(),),
    )

    con.commit()


def stage_partition(
    stream: Stream, partition_values: Optional[List[Any]] = None, *args, **kwargs
) -> Partition:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    partition_id = uuid.uuid4().__str__()
    partition_locator = PartitionLocator.of(
        stream.locator, partition_values=partition_values, partition_id=partition_id
    )

    tv = get_table_version(
        stream.namespace, stream.table_name, stream.table_version, *args, **kwargs
    )

    pv_partition = get_partition(
        stream.locator, partition_values=partition_values, *args, **kwargs
    )

    stream_position = current_time_ms()
    partition = Partition.of(
        partition_locator,
        schema=tv.schema,
        content_types=tv.content_types,
        state=CommitState.STAGED,
        previous_stream_position=pv_partition.stream_position if pv_partition else None,
        previous_partition_id=pv_partition.partition_id if pv_partition else None,
        stream_position=stream_position,
    )

    params = (
        partition.locator.canonical_string(),
        partition.stream_locator.canonical_string(),
        json.dumps(partition),
    )
    cur.execute("INSERT INTO partitions VALUES (?, ?, ?)", params)
    con.commit()

    return partition


def commit_partition(partition: Partition, *args, **kwargs) -> Partition:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    pv_partition = get_partition(
        partition.stream_locator,
        partition_values=partition.partition_values,
        *args,
        **kwargs,
    )

    # deprecate old and commit new one
    if pv_partition:
        pv_partition.state = CommitState.DEPRECATED
        params = (json.dumps(pv_partition), pv_partition.locator.canonical_string())
        cur.execute("UPDATE partitions SET value = ? WHERE locator = ?", params)

    deltas = list_partition_deltas(partition, *args, **kwargs).all_items()
    deltas.sort(reverse=True, key=lambda x: x.stream_position)

    stream_position = partition.stream_position
    if deltas:
        stream_position = deltas[0].stream_position

    partition.state = CommitState.COMMITTED
    partition.stream_position = stream_position
    partition.previous_stream_position = (
        pv_partition.stream_position if pv_partition else None
    )
    params = (json.dumps(partition), partition.locator.canonical_string())
    cur.execute("UPDATE partitions SET value = ? WHERE locator = ?", params)
    con.commit()

    return partition


def delete_partition(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    partition_values: Optional[List[Any]] = None,
    *args,
    **kwargs,
) -> None:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    stream = get_stream(namespace, table_name, table_version, *args, **kwargs)
    partition = get_partition(stream.locator, partition_values, *args, **kwargs)

    partition.state = CommitState.DEPRECATED
    params = (json.dumps(partition), partition.locator.canonical_string())

    cur.execute("UPDATE partitions SET value = ? WHERE locator = ?", params)
    con.commit()


def get_partition(
    stream_locator: StreamLocator,
    partition_values: Optional[List[Any]] = None,
    *args,
    **kwargs,
) -> Optional[Partition]:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    res = cur.execute(
        "SELECT * FROM partitions WHERE stream_locator = ?",
        (stream_locator.canonical_string(),),
    )

    serialized_partitions = res.fetchall()

    if not serialized_partitions:
        return None

    if partition_values is None:
        partition_values = []

    prior_pv = ",".join(partition_values)

    for item in serialized_partitions:
        partition = Partition(json.loads(item[2]))
        pv = ",".join(partition.partition_values if partition.partition_values else [])

        if pv == prior_pv and partition.state == CommitState.COMMITTED:
            return partition

    return None


def stage_delta(
    data: Union[LocalTable, LocalDataset, DistributedDataset, Manifest],
    partition: Partition,
    delta_type: DeltaType = DeltaType.UPSERT,
    max_records_per_entry: Optional[int] = None,
    author: Optional[ManifestAuthor] = None,
    properties: Optional[Dict[str, str]] = None,
    s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
    content_type: ContentType = ContentType.PARQUET,
    *args,
    **kwargs,
) -> Delta:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    manifest_entry_id = uuid.uuid4().__str__()
    uri = _get_manifest_entry_uri(manifest_entry_id)

    serialized_data = None
    if content_type == ContentType.PARQUET:
        buffer = io.BytesIO()
        pa.parquet.write_table(data, buffer)
        serialized_data = buffer.getvalue()
    elif content_type == ContentType.UNESCAPED_TSV:
        buffer = io.BytesIO()
        write_options = pa.csv.WriteOptions(
            include_header=True, delimiter="\t", quoting_style="none"
        )
        pa.csv.write_csv(data, buffer, write_options=write_options)
        serialized_data = buffer.getvalue()
    else:
        raise ValueError(f"Unsupported content type: {content_type}")

    stream_position = current_time_ms()
    delta_locator = DeltaLocator.of(partition.locator, stream_position=stream_position)

    meta = ManifestMeta.of(
        len(data),
        len(serialized_data),
        content_type=content_type,
        content_encoding=ContentEncoding.IDENTITY,
        source_content_length=data.nbytes,
    )

    manifest = Manifest.of(
        entries=ManifestEntryList.of(
            [
                ManifestEntry.of(
                    uri=uri, url=uri, meta=meta, mandatory=True, uuid=manifest_entry_id
                )
            ]
        ),
        author=author,
        uuid=manifest_entry_id,
    )

    delta = Delta.of(
        delta_locator,
        delta_type=delta_type,
        meta=meta,
        properties=properties,
        manifest=manifest,
        previous_stream_position=partition.stream_position,
    )

    params = (uri, serialized_data)
    cur.execute("INSERT OR IGNORE INTO data VALUES (?, ?)", params)

    params = (delta_locator.canonical_string(), "staged_delta", json.dumps(delta))
    cur.execute("INSERT OR IGNORE INTO deltas VALUES (?, ?, ?)", params)

    con.commit()
    return delta


def commit_delta(delta: Delta, *args, **kwargs) -> Delta:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    if not delta.stream_position:
        delta.locator.stream_position = current_time_ms()

    params = (
        delta.locator.canonical_string(),
        delta.partition_locator.canonical_string(),
        json.dumps(delta),
    )

    cur.execute("INSERT OR IGNORE INTO deltas VALUES (?, ?, ?)", params)

    params = (
        delta.partition_locator.canonical_string(),
        json.dumps(delta),
        delta.locator.canonical_string(),
    )
    cur.execute(
        "UPDATE deltas SET partition_locator = ?, value = ? WHERE locator = ?", params
    )

    con.commit()

    return delta


def get_namespace(namespace: str, *args, **kwargs) -> Optional[Namespace]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = NamespaceLocator.of(namespace)

    res = cur.execute(
        "SELECT * FROM namespaces WHERE locator = ?", (locator.canonical_string(),)
    )
    serialized_result = res.fetchone()

    if serialized_result is None:
        return None

    return Namespace(json.loads(serialized_result[1]))


def namespace_exists(namespace: str, *args, **kwargs) -> bool:
    obj = get_namespace(namespace, *args, **kwargs)

    return obj is not None


def get_table(namespace: str, table_name: str, *args, **kwargs) -> Optional[Table]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)

    res = cur.execute(
        "SELECT * FROM tables WHERE locator = ?", (locator.canonical_string(),)
    )
    serialized_result = res.fetchone()

    if serialized_result is None:
        return None

    return Table(json.loads(serialized_result[2]))


def table_exists(namespace: str, table_name: str, *args, **kwargs) -> bool:
    obj = get_table(namespace, table_name, *args, **kwargs)

    return obj is not None


def get_table_version(
    namespace: str, table_name: str, table_version: str, *args, **kwargs
) -> Optional[TableVersion]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = TableVersionLocator.of(
        TableLocator.of(NamespaceLocator.of(namespace), table_name), table_version
    )

    res = cur.execute(
        "SELECT * FROM table_versions WHERE locator = ?", (locator.canonical_string(),)
    )
    serialized_table_version = res.fetchone()

    if serialized_table_version is None:
        return None

    return TableVersion(json.loads(serialized_table_version[2]))


def get_latest_table_version(
    namespace: str, table_name: str, *args, **kwargs
) -> Optional[TableVersion]:
    table_versions = list_table_versions(
        namespace, table_name, *args, **kwargs
    ).all_items()
    if not table_versions:
        return None

    table_versions.sort(reverse=True, key=lambda v: int(v.table_version))
    return table_versions[0]


def get_latest_active_table_version(
    namespace: str, table_name: str, *args, **kwargs
) -> Optional[TableVersion]:

    # This module does not support table version lifecycle state
    return get_latest_table_version(namespace, table_name, *args, **kwargs)


def get_table_version_schema(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Union[pa.Schema, str, bytes]]:
    obj = get_table_version(namespace, table_name, table_version, *args, **kwargs)

    return obj.schema


def table_version_exists(
    namespace: str, table_name: str, table_version: str, *args, **kwargs
) -> bool:
    obj = get_table_version(namespace, table_name, table_version, *args, **kwargs)

    return obj is not None


def get_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Stream]:
    assert not isinstance(table_version, int), f"Passed an integer as the table version"
    obj = get_table_version(namespace, table_name, table_version, *args, **kwargs)

    if obj is None:
        return None

    stream_id = obj.properties.get(STREAM_ID_PROPERTY)
    if stream_id is None:
        return None

    cur, con = _get_sqlite3_cursor_con(kwargs)
    stream_locator = StreamLocator.of(
        obj.locator, stream_id=stream_id, storage_type=STORAGE_TYPE
    )
    res = cur.execute(
        "SELECT * FROM streams WHERE locator = ?", (stream_locator.canonical_string(),)
    )

    serialized_stream = res.fetchone()
    if serialized_stream is None:
        return None

    return Stream(json.loads(serialized_stream[2]))


def get_table_version_column_names(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[List[str]]:
    raise NotImplementedError("Fetching column names is not supported")
